class AtomicIntervalException(Exception):
    pass


class AtomicIntervalWarning(Warning):
    pass


class IntervalsException(Exception):
    pass


class IntervalsWarning(Warning):
    pass


__all__ = [
    'AtomicIntervalException',
    'AtomicIntervalWarning',
    'IntervalsException',
    'IntervalsWarning',
]


import enum
import typing as tp

from abc import ABCMeta, abstractmethod

from .patterns import _Singleton


class _Comparable(metaclass=ABCMeta):
    @abstractmethod
    def __lt__(self, other: tp.Any) -> bool: ...


CT = tp.TypeVar('CT', bound=_Comparable)


class Bound(enum.Enum):
    """
    Bound types.
    """
    OPEN = False
    CLOSED = True

    def bracket(self, is_left: bool = True) -> str:
        if self.value:
            return "[" if is_left else "]"

        return "(" if is_left else ")"

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class _PositiveInf(_Singleton):
    """
    Positive infinity.
    """
    def __neg__(self) -> '_NegativeInf':
        return _NegativeInf()

    def __lt__(self, other: CT) -> bool:
        return False

    def __le__(self, other: CT) -> bool:
        return isinstance(other, _PositiveInf)

    def __gt__(self, other: CT) -> bool:
        return True

    def __ge__(self, other: CT) -> bool:
        return not isinstance(other, _PositiveInf)

    def __eq__(self, other: CT) -> bool:
        return isinstance(other, _PositiveInf)

    def __repr__(self):
        return "inf"

    def __add__(self, other: CT) -> '_PositiveInf':
        if isinstance(other, _NegativeInf):
            raise ValueError("Undefined sum of positive and negative infinities")

        return self

    def __sub__(self, other: CT) -> '_PositiveInf':
        if isinstance(other, _PositiveInf):
            raise ValueError("Undefined difference of positive and negative infinities")

        return self

    def __hash__(self):
        return hash(float("inf"))


class _NegativeInf(_Singleton):
    """
    Negative infinity
    """

    def __neg__(self) -> '_PositiveInf':
        return _PositiveInf()

    def __lt__(self, other: CT) -> bool:
        return True

    def __le__(self, other: CT) -> bool:
        return not isinstance(other, _NegativeInf)

    def __gt__(self, other: CT) -> bool:
        return False

    def __ge__(self, other: CT) -> bool:
        return isinstance(other, _NegativeInf)

    def __eq__(self, other: CT) -> bool:
        return isinstance(other, _NegativeInf)

    def __repr__(self):
        return "-inf"

    def __add__(self, other: CT) -> '_NegativeInf':
        if isinstance(other, _PositiveInf):
            raise ValueError("Undefined sum of positive and negative infinities")

        return self

    def __sub__(self, other: CT) -> '_NegativeInf':
        if isinstance(other, _NegativeInf):
            raise ValueError("Undefined difference of negative and positive infinities")

        return self

    def __hash__(self):
        return hash(float("-inf"))


def convert_json_value(value: tp.Any) -> tp.Any:
    if value == "inf":
        return inf
    elif value == "-inf":
        return -inf
    else:
        return value


inf = _PositiveInf()


__all__ = ['CT', 'Bound', 'inf', 'convert_json_value']


import numpy as np
import typing as tp
import plotly.figure_factory as ff
import heapq

from src.pintle.atomic_interval import *
from src.pintle.bounds import inf, Bound, CT
from src.pintle.exception import IntervalsException
from src.pintle.colors.colors import generate_colors
from src.pintle.algorithms.binary_search_interval import binary_search_interval


class Intervals:
    """
    """
    def __init__(self, *intervals):
        self._intervals: tp.List[AtomicInterval] = []

    @classmethod
    def from_atomic(cls, *args: tp.Union[Bound, CT]) -> 'Intervals':
        instance = cls()
        atomic_interval = AtomicInterval(*args)

        if atomic_interval.empty:
            return instance

        instance._intervals = [atomic_interval]
        return instance

    @classmethod
    def from_atomics(cls, intervals: tp.List[AtomicInterval]) -> 'Intervals':
        instance = cls()

        intervals = [interval for interval in intervals if not interval.empty]
        intervals = sorted(intervals, key=lambda x: x.left_value)
        cur_interval = AtomicInterval()

        for interval in intervals:
            if interval.empty:
                continue

            if cur_interval.empty:
                cur_interval = interval

            elif cur_interval.mergeable(interval):
                cur_interval = cur_interval | interval

            else:
                instance._intervals.append(cur_interval)
                cur_interval = interval

        if cur_interval.empty:
            return instance

        instance._intervals.append(cur_interval)
        return instance

    @classmethod
    def open(cls, left_value: CT, right_value: CT) -> 'Intervals':
        """
        Creates atomic interval: (left_value, right_value)
        """
        return cls().from_atomic(Bound.OPEN, left_value, right_value, Bound.OPEN)

    @classmethod
    def open_closed(cls, left_value: CT, right_value: CT) -> 'Intervals':
        """
        Creates atomic interval: (left_value, right_value]
        """
        return cls().from_atomic(Bound.OPEN, left_value, right_value, Bound.CLOSED)

    @classmethod
    def closed_open(cls, left_value: CT, right_value: CT) -> 'Intervals':
        """
        Creates atomic interval: [left_value, right_value)
        """
        return cls().from_atomic(Bound.CLOSED, left_value, right_value, Bound.OPEN)

    @classmethod
    def closed(cls, left_value: CT, right_value: CT) -> 'Intervals':
        """
        Creates atomic interval: [left_value, right_value]
        """
        return cls().from_atomic(Bound.CLOSED, left_value, right_value, Bound.CLOSED)

    @classmethod
    def point(cls, value: CT) -> 'Intervals':
        """
        Creates one point interval: [value, value]
        """
        return cls().from_atomic(value)

    @classmethod
    def empty_interval(cls):
        """
        Creates empty interval
        """
        return cls().from_atomic()

    @classmethod
    def from_json(cls, json_obj: tp.Any) -> 'Intervals':
        """
        Create intervals from json
        :param json_obj: json object
        :return: intervals
        """
        if isinstance(json_obj, dict):
            return cls.from_atomics(
                [AtomicInterval.from_json(json_obj)]
            )
        elif isinstance(json_obj, list):
            return cls.from_atomics([AtomicInterval.from_json(x) for x in json_obj])
        else:
            raise  IntervalsException("Wrong value type, json_obj must be list or dict")


    @property
    def is_extended(self) -> bool:
        f"""
        Check if {inf} or {-inf} is inside intervals.
        :return: True if {inf} or {-inf} is inside atomic interval, False otherwise.
        """
        if self.empty:
            return False

        return self._intervals[0].is_extended or self._intervals[-1].is_extended

    @property
    def empty(self) -> bool:
        return len(self._intervals) == 0 or all([interval.empty for interval in self._intervals])

    @property
    def atomic(self) -> bool:
        return len(self._intervals) <= 1

    @property
    def length(self) -> float:
        return sum(interval.length for interval in self._intervals)

    @property
    def atomic_intervals(self) -> tp.List[AtomicInterval]:
        return self._intervals

    @property
    def max_bound(self) -> tp.Optional[CT]:
        if self.empty:
            return None

        return self._intervals[-1].right_value

    @property
    def min_bound(self) -> tp.Optional[CT]:
        if self.empty:
            return None

        return self._intervals[0].left_value

    def closure(self, is_extended: tp.Optional[bool] = None) -> 'Intervals':
        f"""
        The closure of a subset is the smallest closed subset. 
        Return closure of {self}.
        :param is_extended: If {None}, set this value to {self.is_extended};
                            If True, finds closure in the extended space [-inf, inf]; 
                            If False, finds closure in the space (-inf, inf).
        :return: closure of given intervals.
        """
        if self.empty:
            return self.__class__.from_atomics([])

        is_extended = self.is_extended if is_extended is None else is_extended

        closure_intervals: tp.List[AtomicInterval] = []

        iterator: tp.Iterator[AtomicInterval] = iter(self)
        cur_interval: AtomicInterval = next(iterator).closure(is_extended)

        for interval in self._intervals:
            if interval.adjacent(cur_interval):
                cur_interval = (cur_interval + interval).closure(is_extended)
            else:
                closure_intervals.append(cur_interval)
                cur_interval = interval.closure(is_extended)

        closure_intervals.append(cur_interval)
        return self.__class__.from_atomics(closure_intervals)

    def interior(self) -> 'Intervlas':
        f"""
        The interior of intervals is equal to the maximum open set in {self}.
        :return: interior of {self}
        """
        return self.__class__.from_atomics(
            [interval.interior() for interval in self]
        )

    def boundary(self, is_extended: tp.Optional[bool]) -> 'Intervals':
        """
        The boundary of intervals is equal to the boundary of intervals complement.
        :param is_extended: If {None}, set this value to {self.is_extended};
                            If True, finds closure in the extended space [-inf, inf];
                            If False, finds closure in the space (-inf, inf).
        :return: intervals boundary
        """
        if self.empty:
            return self.__class__.from_atomics([])

        boundary_intervals: tp.List[AtomicInterval] = []
        is_extended = self.is_extended if is_extended is None else is_extended
        iterator: tp.Iterator[AtomicInterval] = iter(self)

        for interval in iterator:
            b_intervals = interval.boundary(is_extended)
            if not boundary_intervals:
                boundary_intervals = b_intervals
            else:
                for b_interval in b_intervals:
                    if boundary_intervals[-1] != b_interval:
                        boundary_intervals.append(b_interval)

        return self.__class__.from_atomics(boundary_intervals)

    def trim(self, value: CT, to_right: bool = True):
        """
        Return trimmed intervals.
        :param value: trimmed boarder.
        :param to_right: the direction of trimming.
        :return: trimmed interval.
        """
        if value <= self._intervals[0].right_value:
            index = 0
        elif value >= self._intervals[-1].right_value:
            index = len(self._intervals) - 1
        else:
            l_pointer, r_pointer = 0, len(self._intervals) - 1

            while r_pointer - l_pointer > 0:
                m_pointer = (r_pointer + l_pointer) // 2
                if self._intervals[m_pointer].right_value >= value:
                    r_pointer = m_pointer
                else:
                    l_pointer = m_pointer + 1

            index = r_pointer

        if to_right:
            intervals = [self._intervals[index].trim(value, to_right)] + self._intervals[index + 1:]
        else:
            intervals = self._intervals[:index] + [self._intervals[index].trim(value, to_right)]

        return self.__class__.from_atomics(intervals)

    def apply(self,
              left_bound: tp.Callable[[Bound], Bound] = None,
              left_value: tp.Callable[[CT], CT] = None,
              right_value: tp.Callable[[CT], CT] = None,
              right_bound: tp.Callable[[Bound], Bound] = None,
              ignore_inf: bool = True, ) -> 'Intervals':
        f"""
        Apply any function to all intervals values. 
        """
        return Intervals.from_atomics(
            [
                interval.apply(
                    left_bound=left_bound,
                    left_value=left_value,
                    right_value=right_value,
                    right_bound=right_bound,
                    ignore_inf=ignore_inf) for interval in self._intervals
            ]
        )

    def to_json(self) -> tp.List[tp.Mapping[str, tp.Union[str, CT]]]:
        """
        Convert intervals to json array
        """
        return [interval.to_json() for interval in self._intervals]

    def adjacent(self, other: 'Intervals') -> bool:
        """
        2 intervals are adjacent if they aren't intersected and their union is atomic interval.
        For example [0,1), [2,3) and [1, 2) are adjacent intervals.
        :param other: other interval.
        :return: True if intervals are adjacent, False otherwise.
        """
        if not isinstance(other, Intervals):
            raise IntervalsException(f"Unsupported type {type(other)} for {other}!")

        if self.empty or other.empty:
            return other.atomic if self.empty else self.atomic

        if self.intersected(other):
            return False

        iterators: tp.List[tp.Iterator[AtomicInterval]] = [iter(self), iter(other)]

        cur_interval = AtomicInterval()
        heap: tp.List[tp.Tuple[float, AtomicInterval, int]] = Intervals._init_heap(*iterators)

        while len(heap) > 0:
            try:
                _, interval, index = heapq.heappop(heap)

                if cur_interval.empty:
                    cur_interval = interval

                elif cur_interval.adjacent(interval):
                    cur_interval = cur_interval | interval
                else:
                    return False

                next_interval = next(iterators[index])
                heapq.heappush(heap, (next_interval.left_value, next_interval, index))
            except StopIteration:
                pass

        return True

    def intersected(self, other: 'Intervals') -> bool:
        """
        Check if 2 intervals are intersected.
        :param other: other interval
        :return: True if intervals are intersected, False otherwise
        """
        if not isinstance(other, Intervals):
            raise IntervalsException(f"Unsupported type {type(other)} for {other}!")

        if self.empty or other.empty:
            return False

        iterators: tp.List[tp.Iterator[AtomicInterval]] = [iter(self), iter(other)]

        cur_interval = AtomicInterval()
        heap: tp.List[tp.Tuple[float, AtomicInterval, int]] = Intervals._init_heap(*iterators)

        while len(heap) > 0:
            try:
                _, interval, index = heapq.heappop(heap)

                if cur_interval.intersected(interval):
                    return True

                if interval not in cur_interval:
                    cur_interval = interval

                next_interval = next(iterators[index])
                heapq.heappush(heap, (next_interval.left_value, next_interval, index))
            except StopIteration:
                pass

        return False

    def complement(self, other: tp.Optional['Intervals'] = None) -> 'Intervals':
        f"""
        Construct complement to {other} interval. 
        If {other} interval is None, the complement to full set would be constructed. 
        :param other: other interval
        :return: complement to other interval
        """
        if other is None:
            other = self.__class__.open(-inf, inf)

        if not isinstance(other, Intervals):
            raise IntervalsException(f"Unsupported type {type(other)} for {other}!")

        if self.empty:
            return other

        if self._intervals[0].left_value == -inf:
            self._intervals[0].apply(left_bound=lambda _: Bound.OPEN)

        if self._intervals[-1].right_value == inf:
            self._intervals[-1].apply(right_bound=lambda _: Bound.OPEN)

        if self not in other:
            raise IntervalsException(f"""
                It is impossible to construct the complement because the interval {self} is not in {other}
            """)

        result = []
        self_iterator, other_iterator = iter(self), iter(other)
        other_atomic = next(other_iterator)

        for self_atomic in self_iterator:
            while self_atomic not in other_atomic:
                result.append(other_atomic)
                other_atomic = next(other_iterator)

            result.append(
                AtomicInterval(
                    other_atomic.left_bound,
                    other_atomic.left_value,
                    self_atomic.left_value,
                    Bound.OPEN if self_atomic.left_bound == Bound.CLOSED else Bound.CLOSED,
                )
            )

            other_atomic = AtomicInterval(
                Bound.OPEN if self_atomic.right_bound == Bound.CLOSED else Bound.CLOSED,
                self_atomic.right_value,
                other_atomic.right_value,
                other_atomic.right_bound,
            )

        result.append(other_atomic)
        for other_atomic in other_iterator:
            result.append(other_atomic)

        return self.__class__.from_atomics(result)

    def difference(self, other: 'Intervals') -> 'Intervals':
        # todo: faster?
        f"""
        Construct the difference between {self} interval and {other} interval.  
        :param other: other interval
        :return: Difference intervals 
        """
        if not isinstance(other, Intervals):
            raise IntervalsException(f"Unsupported type {type(other)} for {other}!")

        return (self & other).complement(self)

    def _contains_atomic(self, other: AtomicInterval) -> bool:
        f"""
        Test if atomic {other} interval is inside intervals {self._intervals}. 
        :param other:
        :return:
        """
        return binary_search_interval(self._intervals, other)

    @staticmethod
    def _init_heap(*iterators: tp.Iterator[AtomicInterval]) -> tp.List[tp.Tuple[float, AtomicInterval, int]]:
        """
        Heap initialization via adding the first elements of each iterator
        """
        heap: tp.List[tp.Tuple[float, AtomicInterval, int]] = []

        for i, iterator in enumerate(iterators):
            atomic_interval = next(iterator)
            heapq.heappush(heap, (atomic_interval.left_value, atomic_interval, i))

        return heap

    def __iter__(self):
        yield from self._intervals

    def __contains__(self, other: tp.Union['Intervals', CT]) -> bool:
        if not isinstance(other, Intervals):
            try:
                other = Intervals.from_atomic(other)
            except:
                raise IntervalsException(f"Unsupported type {type(other)} for {other}!")

        iterator: tp.Iterator[AtomicInterval] = iter(self)
        cur_interval: AtomicInterval = next(iterator)

        for search_interval in other.atomic_intervals:
            try:
                if search_interval in cur_interval:
                    continue

                if cur_interval > search_interval:
                    return False
                cur_interval = next(iterator)
            except StopIteration:
                return False

        return True

    def __len__(self) -> int:
        return len(self._intervals)

    def __eq__(self, other: 'Intervals') -> bool:
        return self._intervals == other._intervals

    def __repr__(self) -> str:
        return ",".join([repr(interval) for interval in self._intervals])

    def __and__(self, other: 'Intervals') -> 'Intervals':
        if self.empty or other.empty:
            return Intervals.from_atomic()

        result_interval = Intervals.from_atomic()

        iterators: tp.List[tp.Iterator[AtomicInterval]] = [iter(self), iter(other)]

        cur_interval = AtomicInterval()
        heap: tp.List[tp.Tuple[float, AtomicInterval, int]] = Intervals._init_heap(*iterators)

        while len(heap) > 0:
            try:
                _, interval, index = heapq.heappop(heap)

                if cur_interval.intersected(interval):
                    result_interval._intervals.append(cur_interval & interval)

                if interval not in cur_interval:
                    cur_interval = interval

                next_interval = next(iterators[index])
                heapq.heappush(heap, (next_interval.left_value, next_interval, index))
            except StopIteration:
                pass

        return result_interval

    def __or__(self, other: 'Intervals') -> 'Intervals':
        if self.empty:
            return other

        if other.empty:
            return self

        result_interval = Intervals.from_atomic()
        iterators: tp.List[tp.Iterator[AtomicInterval]] = [iter(self), iter(other)]

        cur_interval = AtomicInterval()
        heap: tp.List[tp.Tuple[float, AtomicInterval, int]] = Intervals._init_heap(*iterators)

        while len(heap) > 0:
            try:
                _, interval, index = heapq.heappop(heap)

                if cur_interval.empty:
                    cur_interval = interval

                elif cur_interval.mergeable(interval):
                    cur_interval = cur_interval | interval

                else:
                    result_interval._intervals.append(cur_interval)
                    cur_interval = interval

                next_interval = next(iterators[index])
                heapq.heappush(heap, (next_interval.left_value, next_interval, index))
            except StopIteration:
                pass

        if not cur_interval.empty:
            result_interval._intervals.append(cur_interval)
        return result_interval

    def __add__(self, other: 'Intervals') -> 'Intervals':
        return self | other

    def __sub__(self, other: 'Intervals') -> 'Intervals':
        return self & other

    def visualize(self, **kwargs):
        # todo: for different types of data - different visualization
        if self.empty:
            raise IntervalsException("Can't visualize empty intervals")

        left_value, right_value = 0, 0
        right_text, left_text = left_value, right_value
        data: tp.List[tp.Mapping[str, tp.Any]] = []

        for i, interval in enumerate(self._intervals):
            if interval.left_value == -inf and interval.right_value == inf:
                data.append(
                    dict(
                        Task="Interval",
                        Start=0,
                        Finish=1,
                        Interval=str(interval)
                    )
                )
                left_value, right_value = 0, 1
                right_text, left_text = '-inf', 'inf'

            elif interval.left_value == -inf:
                data.append(
                    dict(
                        Task="Interval",
                        Start=interval.right_value,
                        Finish=interval.right_value - 1,
                        Interval=str(interval)
                    )
                )
                left_value, left_text = interval.right_value - 1, '-inf'

            elif interval.right_value == inf:
                data.append(
                    dict(
                        Task="Interval",
                        Start=interval.left_value,
                        Finish=interval.left_value + 1,
                        Interval=str(interval)
                    )
                )
                right_value, right_text = interval.left_value + 1, 'inf'

            else:
                data.append(
                    dict(
                        Task="Interval",
                        Start=interval.left_value,
                        Finish=interval.right_value,
                        Interval=str(interval)
                    )
                )
                if i == 0:
                    left_value, left_text = interval.left_value, interval.left_value

                if i == len(self) - 1:
                    right_value, right_text = interval.right_value, interval.right_value

        colors = generate_colors(len(self))
        fig = ff.create_gantt(
            data,
            colors=dict(zip([str(interval) for interval in self], colors)),
            index_col='Interval',
            group_tasks=True,
            height=250,
            title=f'Atomic interval {str(self)}',
            showgrid_x=True,
            bar_width=0.5
        )

        annotations = [
            dict(x=(d['Start'] + d['Finish']) / 2, y=0, text=d['Interval'], showarrow=False, font=dict(color='black'))
            for d in data
        ]
        fig['layout']['annotations'] = annotations

        # todo: xaxis type
        tick_values = list(round(x) for x in np.linspace(left_value, right_value, 10))
        tick_text = tick_values.copy()
        tick_text[0] = left_text
        tick_text[-1] = right_text

        fig.update_layout(
            xaxis_type='linear',
            xaxis=dict(
                tickmode='array',
                tickvals=tick_values,
                ticktext=tick_text,
            ),
            yaxis_visible=False,
        )
        return fig



import seaborn as sns
import typing as tp


def generate_colors(n: int, style: str = "pastel") -> tp.List[str]:
    """
    Generate random n colors
    """
    return sns.color_palette(style, n)


__all__ = ['generate_colors']


class _Singleton:
    """
    Singleton pattern realization.
    """
    __instance = None

    def __new__(cls, *args, **kwargs):
        if not cls.__instance:
            cls.__instance = super().__new__(cls)

        return cls.__instance


__all__ = ['_Singleton']


import typing as tp

import json
import warnings
import plotly.figure_factory as ff

from src.pintle.bounds import *
from src.pintle.exception import AtomicIntervalException, AtomicIntervalWarning
from src.pintle.colors.colors import generate_colors


class AtomicInterval:
    def __init__(self, *args: tp.Union[Bound, CT, None]):
        """
        Atomic interval is simple interval with 2 types of bounds. Each bound is either open or closed.
        It is worth mentioning that one point is an atomic interval
        """
        assert len(args) in [0, 1, 2, 4], \
            AtomicIntervalException(f"Wrong number of arguments. The number of arguments should be 0, 1, 2 or 4")

        if len(args) == 0:
            self._init_empty_interval()

        elif len(args) == 1:
            self._init_interval(Bound.CLOSED, args[0], args[0], Bound.CLOSED)

        elif len(args) == 2:
            if args[0] > args[1]:
                warnings.warn(
                    f"""
                    Empty interval created as the left value of the interval {args[0]} 
                    is greater than the right value {args[1]} of the interval""",
                    AtomicIntervalWarning
                )
                self._init_empty_interval()
            else:
                self._init_interval(Bound.OPEN, *args, Bound.OPEN)

        else:
            if args[1] > args[2]:
                warnings.warn(
                    f"""
                    Empty interval created as the left value of the interval {args[0]} 
                    is greater than the right value {args[1]} of the interval""",
                    AtomicIntervalWarning
                )
                self._init_empty_interval()

            else:
                self._init_interval(*args)

                if self._left_value == self.right_value and \
                   (self._left_bound == Bound.OPEN or self._right_bound == Bound.OPEN):

                    self._init_empty_interval()

    def _init_empty_interval(self) -> None:
        """
        Init params of empty interval
        """
        self._left_value, self._right_value = None, None
        self._left_bound, self._right_bound = Bound.OPEN, Bound.OPEN

    def _init_interval(self, *args) -> None:
        """
        Init params of not empty interval
        """
        self._left_bound, self._left_value, self._right_value, self._right_bound = args

    def __repr__(self) -> str:
        """
        Returns the atomic interval representation. For example [0, 1), [2, +inf) or [3]
        """
        left_bracket = self._left_bound.bracket(is_left=True)
        right_bracket = self._right_bound.bracket(is_left=False)

        if self._left_value == self._right_value:
            return f"{left_bracket}{self._left_value}{right_bracket}"

        return f"{left_bracket}{self._left_value}, {self._right_value}{right_bracket}"

    def mergeable(self, b: 'AtomicInterval') -> bool:
        """
        Check if 2 atomic intervals could be merged.
        """
        if self.empty or b.empty:
            return True

        first, second = sorted((self, b), key=lambda x: x.left_value)

        if first.right_value == second.left_value:
            return first.right_bound == Bound.CLOSED or second.left_bound == Bound.CLOSED

        return first.right_value > second.left_value

    @property
    def is_extended(self) -> bool:
        f"""
        Check if {inf} or {-inf} is inside atomic interval
        :return: True if {inf} or {-inf} is inside atomic interval, False otherwise
        """
        if self.left_bound == Bound.CLOSED and self.left_value == -inf:
            return True

        if self.right_bound == Bound.CLOSED and self.right_value == inf:
            return True

        return False

    @property
    def left_bound(self) -> Bound:
        """
        Left bound of atomic interval
        """
        return self._left_bound

    @property
    def right_bound(self) -> Bound:
        """
        Right bound of atomic interval
        """
        return self._right_bound

    @property
    def left_value(self) -> CT:
        """
        Left value of atomic interval
        """
        return self._left_value

    @property
    def right_value(self) -> CT:
        """
        Right bound of atomic interval
        """
        return self._right_value

    @property
    def empty(self) -> bool:
        """
        Check if interval is empty (doesn't contain any point)
        """
        if (self._left_value is None) or (self._right_value is None):
            return True

        return False

    @property
    def length(self) -> float:
        """
        Length of interval. If interval is [inf, inf] or [-inf, -inf], the property length would return 0
        :return:
        """
        if self.empty:
            return 0

        if hasattr(self._left_value, '__sub__') and hasattr(self._right_value, '__sub__'):
            if self._right_value == inf:
                return inf

            elif self._left_value == -inf:
                return inf

            else:
                return self._right_value - self._left_value
        else:
            raise AtomicIntervalException(
                f"Unsupported operand type(s) for -: {type(self._left_value)} and {type(self._right_value)}"
            )

    def closure(self, is_extended: tp.Optional[bool] = None) -> 'AtomicInterval':
        """
        Closure of atomic interval. For example, (0, 1) -> [0, 1]
        """
        if self.empty:
            return self.__class__()

        is_extended = self.is_extended if is_extended is None else is_extended

        if is_extended:
            left_bound, right_bound = Bound.CLOSED, Bound.CLOSED
        else:
            left_bound: Bound = Bound.CLOSED if self._left_value != -inf else self._left_bound
            right_bound: Bound = Bound.CLOSED if self._right_value != inf else self._right_bound

        return self.__class__(left_bound, self.left_value, self.right_value, right_bound)

    def boundary(self, is_extended: tp.Optional[bool] = None) -> tp.List['AtomicInterval']:
        """
        Returns boundary of atomic interval
        """
        if self.empty:
            return []

        is_extended = self.is_extended if is_extended is None else is_extended
        if is_extended:
            if self._left_value == self._right_value:
                return [self.__class__(self._left_value)]
            else:
                return [self.__class__(self._left_value), self.__class__(self._right_value)]
        else:
            boundary_intervals: tp.List['AtomicInterval'] = []

            if self._left_value != -inf:
                boundary_intervals.append(self.__class__(self._left_value))

            if self._right_value != inf and self._right_value != self.left_value:
                boundary_intervals.append(self.__class__(self._right_value))

            return boundary_intervals

    def interior(self) -> 'AtomicInterval':
        """
        Returns the interior of atomic interval.
        """
        if self._left_value == self._right_value:
            return self.__class__()

        return self.__class__(Bound.OPEN, self._left_value, self._right_value, Bound.OPEN)

    def intersected(self, other: 'AtomicInterval') -> bool:
        """
        Check if two intervals are intersecting
        """
        if self.empty or other.empty:
            return False

        first, second = sorted((self, other), key=lambda x: (x.left_value, x.right_value))

        if first.right_value == second.left_value:
            return first.right_bound == Bound.CLOSED and second.left_bound == Bound.CLOSED

        return first.right_value > second.left_value

    def adjacent(self, other: 'AtomicInterval') -> bool:
        """
        Check if two intervals are not intersected but mergeable.
        """
        if self.empty or other.empty:
            return True

        first, second = sorted((self, other), key=lambda x: (x.left_value, x.right_value))
        if first.right_value == second.left_value:
            return int(first.right_bound == Bound.CLOSED) + int(second.left_bound == Bound.CLOSED) == 1

        return False

    def apply(self,
              left_bound: tp.Callable[[Bound], Bound] = None,
              left_value: tp.Callable[[CT], CT] = None,
              right_value: tp.Callable[[CT], CT] = None,
              right_bound: tp.Callable[[Bound], Bound] = None,
              ignore_inf: bool = True,
              ) -> 'AtomicInterval':
        """
        Apply some function on right and left values of interval. Return new atomic interval
        """
        if self.empty:
            return self

        l_bound = self._left_bound if not left_bound else left_bound(self._left_bound)

        if left_value is not None:
            if ignore_inf and (self._left_value == inf or self._left_value == -inf):
                l_value = self._left_value
            else:
                l_value = left_value(self._left_value)
        else:
            l_value = self._left_value

        if right_value is not None:
            if ignore_inf and (self._right_value == inf or self._right_value == -inf):
                r_value = self._right_value
            else:
                r_value = right_value(self._right_value)
        else:
            r_value = self._right_value

        r_bound = self._right_bound if not right_bound else right_bound(self._right_bound)

        return self.__class__(l_bound, l_value, r_value, r_bound)

    def trim(self, value: CT, to_right=True) -> 'AtomicInterval':
        """
        Cut the atomic interval. For example, if interval is [0, 5], value = 3 and to_right=True,
        the result would be [3, 5].
        """
        if value in self:
            if to_right:
                return AtomicInterval(Bound.CLOSED, value, self._right_value, self._right_bound)
            else:
                return AtomicInterval(self._left_bound, self._left_value, value, Bound.CLOSED)

        if to_right:
            return AtomicInterval() if value >= self._right_value else self
        else:
            return AtomicInterval() if value <= self._left_value else self

    def to_json(self) -> tp.Mapping[str, tp.Union[str, float, int]]:
        """
        #todo: datetime
        Convert  atomic interval to json
        :return: dictionary with interval parameters
        """
        if self.empty:
            return {}
        else:
            return {
                'left_bound': self._left_bound.name,
                'left_value': self._left_value if isinstance(self._left_value, (float, int)) else str(self._left_value),
                'right_value': self._right_value if isinstance(self._right_value, (float, int)) else str(self._right_value),
                'right_bound': self._right_bound.name,
            }

    @classmethod
    def from_json(cls, json_obj: tp.Union[str, tp.Mapping[str, tp.Any]]):
        """
        Build atomic interval from json object
        :param json_obj: json object
        :return: atomic interval
        """
        if isinstance(json_obj, str):
            json_obj = json.loads(json_obj)

        if not isinstance(json_obj, dict):
            raise AtomicIntervalException("""
                Wrong format of json data. Shoud be dictionary with
                - left_bound,
                
                """)

        if not json_obj:
            return cls()
        else:
            left_bound = Bound.CLOSED if json_obj['left_bound'] == 'CLOSED' else Bound.OPEN
            left_value = convert_json_value(json_obj['left_value'])
            right_value = convert_json_value(json_obj['right_value'])
            right_bound = Bound.CLOSED if json_obj['right_bound'] == 'CLOSED' else Bound.OPEN

            return cls(left_bound, left_value, right_value, right_bound)

    def __len__(self) -> int:
        return int(self.empty)

    def __iter__(self) -> 'AtomicInterval':
        yield self

    def __and__(self, other: 'AtomicInterval') -> 'AtomicInterval':
        if not self.intersected(other):
            return AtomicInterval()

        left_value = max(self._left_value, other._left_value)
        right_value = min(self._right_value, other._right_value)

        left_bound, right_bound = Bound.CLOSED, Bound.CLOSED

        for x in [self, other]:
            if x.left_value == left_value and left_bound != Bound.OPEN:
                left_bound = x.left_bound

            if x.right_value == right_value and right_bound != Bound.OPEN:
                right_bound = x.right_bound

        return AtomicInterval(left_bound, left_value, right_value, right_bound)

    def __or__(self, other: 'AtomicInterval') -> 'AtomicInterval':
        if not self.mergeable(other):
            raise AtomicIntervalException("Two intervals can't be merged")

        if self.empty:
            return other

        if other.empty:
            return other

        left_value = min(self._left_value, other._left_value)
        right_value = max(self._right_value, other._right_value)

        left_bound, right_bound = Bound.OPEN, Bound.OPEN

        for x in [self, other]:
            if x.left_value == left_value and left_bound != Bound.CLOSED:
                left_bound = x.left_bound

            if x.right_value == right_value and right_bound != Bound.CLOSED:
                right_bound = x.right_bound

        return AtomicInterval(left_bound, left_value, right_value, right_bound)

    def __contains__(self, other: tp.Union['AtomicInterval', CT]) -> bool:
        # todo: refactor!!!!
        if self.empty:
            return False

        if not isinstance(other, self.__class__):
            other = AtomicInterval(other)

        if self._left_value > other.left_value or self._right_value < other.right_value:
            return False

        if other._left_value == self._left_value and other.left_bound.value > self.left_bound.value:
            return False

        if other._right_value == self._right_value and other.right_bound.value > self.right_bound.value:
            return False

        return True

    def __add__(self, other: 'AtomicInterval') -> 'AtomicInterval':
        return self | other

    def __sub__(self, other):
        return self & other

    def __lt__(self, other: 'AtomicInterval') -> bool:
        return self.right_value <= other.left_value and not self.intersected(other)

    def __eq__(self, other: 'AtomicInterval') -> bool:
        return (self._left_value == other.left_value) and \
            (self._right_value == other.right_value) and \
            (self._left_bound == other.left_bound) and \
            (self._right_bound == other.right_bound)

    def __hash__(self):
        return hash(
            tuple(
                [
                    self.left_bound,
                    self.left_value,
                    self.right_value,
                    self.right_bound,
                ]
            )
        )

    def visualize(self, **kwargs):
        if self.empty:
            raise AtomicIntervalException("Can't visualize empty atomic interval")

        xaxis_dict = {'tickmode': 'array'}

        if self._left_value == -inf and self._right_value == inf:
            interval = (0, 1)
            xaxis_dict.update(
                dict(
                    tickvals=interval,
                    ticktext=('-inf', 'inf'),
                )
            )
        elif self._left_value == -inf:
            interval = (self._right_value - 1, self._right_value)
            xaxis_dict.update(
                dict(
                    tickvals=interval,
                    ticktext=('-inf', self._right_value),
                )
            )
        elif self._right_value == inf:
            interval = (self._left_value, self._left_value + 1)
            xaxis_dict.update(
                dict(
                    tickvals=interval,
                    ticktext=(self._left_value, 'inf'),
                )
            )
        else:
            interval = (self._left_value, self._right_value)

        fig = ff.create_gantt(
            [dict(Task="AtomicInterval", Start=interval[0], Finish=interval[1], Interval=str(self))],
            colors={str(self): generate_colors(1)[0]},
            index_col='Interval',
            group_tasks=True,
            height=250,
            title=f'Atomic interval {str(self)}',
            showgrid_x=True,
            bar_width=0.5
        )
        # todo: xaxis type
        fig.update_layout(
            xaxis_type='linear',
            xaxis=xaxis_dict,
            yaxis_visible=False,
        )
        return fig


__all__ = ['AtomicInterval']


import typing as tp

T = tp.TypeVar('T', bound=tp.Container)


def binary_search_interval(intervals: tp.List[T], item: T) -> bool:
    f"""
    The problem is as follows. There are given set of not overlapped intervals {intervals}.
    These intervals are sorted by beginning. Find out if one of intervals contains {item} interval. 
    """
    if not intervals:
        return False

    l_pointer, m_pointer, r_pointer = 0, 0, len(intervals) - 1

    while l_pointer < r_pointer:
        m_pointer = (l_pointer + r_pointer) // 2

        if intervals[m_pointer] < item:
            l_pointer = m_pointer + 1

        elif intervals[m_pointer] > item:
            r_pointer = m_pointer - 1

        else:
            if item in intervals[m_pointer]:
                return True
            return False

    return item in intervals[l_pointer]
